import os
from collections import defaultdict
from typing import DefaultDict, Dict, Generator, Type
from types import ModuleType
import inspect

from fastapi import FastAPI
from dapr.ext.fastapi import DaprActor

from kess import env, imports, actor
from kess.function import Function


class App(FastAPI):
    production: bool
    functions_folder: str
    runtime_prefix: str
    runtime_name: str
    functions: DefaultDict[str, Dict[str, Function]]
    actors: DefaultDict[str, Dict[str, Type[actor.Actor]]]

    def __init__(self, *args, **kwargs):
        self.functions_folder = kwargs.pop("functions_folder", env.FN_FOLDER)
        self.runtime_prefix = kwargs.pop("runtime_prefix", env.RUNTIME_PREFIX)
        self.runtime_name = kwargs.pop("runtime_name", env.RUNTIME_NAME)
        self.functions = defaultdict(dict)
        self.actors = defaultdict(dict)
        super().__init__(*args, **kwargs)

        actor = DaprActor(self)
        @self.on_event("startup")
        async def _startup():
            for _, versions in self.actors.items():
                for _, a in versions.items():
                    await actor.register_actor(a)

    @property
    def route_prefix(self):
        return f"{self.runtime_prefix}"

    def scan_prod_functions(
        self, functions_folder: str
    ) -> Generator[(str, str, ModuleType)]:
        for function_name in os.listdir(functions_folder):
            function_path = os.path.join(functions_folder, function_name)
            if os.path.isdir(function_path):
                for function_file_name in os.listdir(function_path):
                    function_file_path = os.path.join(function_path, function_file_name)
                    if os.path.isfile(function_file_path):
                        function_version = os.path.splitext(function_file_name)[0]
                        module = imports.import_module(
                            os.path.splitext(function_file_path)[0].replace(
                                os.path.sep, "."
                            )
                        )
                        yield function_name, function_version, module

    def scan_dev_functions(
        self, functions_folder: str
    ) -> Generator[(str, str, ModuleType)]:
        for function_file_name in os.listdir(functions_folder):
            function_file_path = os.path.join(functions_folder, function_file_name)
            if os.path.isfile(function_file_path):
                function_name = os.path.splitext(function_file_name)[0]
                function_version = "latest"
                module = imports.import_module(
                    os.path.splitext(function_file_path)[0].replace(os.path.sep, ".")
                )
                yield function_name, function_version, module

    def scan_functions(
        self, functions_folder: str, production: bool = False
    ) -> Generator[(str, str, Function)]:
        scanner = self.scan_prod_functions if production else self.scan_dev_functions
        yield from scanner(functions_folder)

    def setup_function(
        self, module: ModuleType, name: str, version: str, prefix: str = ""
    ):
        for k, v in module.__dict__.items():
            if k.startswith("_"):
                continue
            if isinstance(v, Function):
                v.state.app = self
                self.mount(f"{prefix}/{name}/{version}", v)
                self.functions[name][version] = v
            elif inspect.isclass(v) and issubclass(v, actor.Actor):
                self.actors[name][version] = actor.create(v)

    def setup_functions(
        self, functions_folder: str, production: bool = True, prefix: str = ""
    ):
        for name, version, module in self.scan_functions(
            functions_folder, production=production
        ):
            self.setup_function(module, name, version, prefix=prefix)

    def setup(self):
        super().setup()
        self.setup_functions(self.functions_folder, prefix=self.route_prefix)



from typing import Type, Dict, Any, Callable, Union

from dapr import actor
from dapr.serializers.json import DefaultJSONSerializer

defaultJSONSerializer = DefaultJSONSerializer()

Remindable = actor.Remindable


class Actor:
    ...


class ActorProxyFactory(actor.ActorProxyFactory):
    def create(
        self,
        actor_type: str,
        actor_id: actor.ActorId,
        actor_interface: Type[actor.ActorInterface] = None,
    ) -> "ActorProxy":
        return ActorProxy(
            self._dapr_client,
            actor_type,
            actor_id,
            actor_interface,
            self._message_serializer,
        )


class ActorProxy(actor.ActorProxy):
    _default_proxy_factory = ActorProxyFactory()

    async def invoke(self, method: str, data: Any = None) -> Any:
        raw_data = self._message_serializer.serialize(data)
        resp = await self.invoke_method(method, raw_data)
        return self._message_serializer.deserialize(resp)


def method(func_or_name: Union[Callable, str]):
    if callable(func_or_name):
        func_or_name.__actormethod__ = func_or_name.__name__
        print(func_or_name.__actormethod__)
        return func_or_name

    def wrapper(funcobj):
        funcobj.__actormethod__ = func_or_name
        return funcobj

    return wrapper


def create_interface(obj: Type[Actor], name: str = None) -> Type[actor.ActorInterface]:
    return create_interface_from_dict(dict(obj.__dict__), name or obj.__name__)


def create_interface_from_dict(obj: Dict, name: str) -> Type[actor.ActorInterface]:
    injects = {
        k: v
        for k, v in obj.items()
        if not k.startswith("_") and callable(v) and hasattr(v, "__actormethod__")
    }
    return type(name, (actor.ActorInterface,), injects)


def create(obj: Type[Actor], name: str = None) -> Type[actor.Actor]:
    return create_from_dict(dict(obj.__dict__), name or obj.__name__)


def create_from_dict(obj: Dict, name: str) -> Type[actor.Actor]:
    Interface = create_interface_from_dict(obj, name)
    return type(name, (actor.Actor, Interface), obj)


def create_proxy(
    obj: Type[Actor] = None, name: str = None, id: str = None
) -> ActorProxy:
    if obj:
        return create_proxy_from_dict(dict(obj.__dict__), name or obj.__name__, id=id)
    return create_proxy_from_none(name, id=id)


def create_proxy_from_dict(obj: Dict, name: str, id: str = None) -> ActorProxy:
    actor_id = actor.ActorId(id) if id else actor.ActorId.create_random_id()
    Interface = create_interface_from_dict(obj, name=name)
    return ActorProxy.create(name, actor_id, Interface)


def create_proxy_from_none(name: str, id: str = None) -> ActorProxy:
    actor_id = actor.ActorId(id) if id else actor.ActorId.create_random_id()
    return ActorProxy.create(name, actor_id)


async def invoke(actor: str, method: str, id: str = None, data: Any = None) -> Any:
    return await create_proxy(name=actor).invoke(method, data)



from kess.app import App
from kess.function import Function
from kess.exceptions import *
from kess.actor import (
    Actor,
    Remindable,
    method as actormethod,
    create_interface as ActorInterface,
    create_proxy as ActorProxy,
    invoke as actorinvoke,
)



import httpx


class InternalRPC:
    _runtime: str
    _fn: str
    _version: str

    def __init__(self, runtime=None, fn=None, version="v1"):
        self._runtime = runtime
        self._fn = fn
        self._version = version

    def runtime(self, name):
        self._runtime = name
        return self

    def rt(self, name):
        return self.runtime(name)

    def fn(self, name, version="v1"):
        self._fn = name
        self._version = version
        return self

    @property
    def base_url(self):
        return f"http://{self._runtime}"

    def url(self, path):
        return f"/{self._fn}/{self._version}{path}"

    async def call(self, **kwargs):
        method = kwargs.pop("method", "post")
        path = kwargs.pop("path", "")
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            resp = await getattr(client, method)(self.url(path), **kwargs)
            return resp.json()

    async def post(self, **kwargs):
        kwargs["method"] = "post"
        return await self.call(**kwargs)

    async def get(self, **kwargs):
        kwargs["method"] = "get"
        return await self.call(**kwargs)

    async def __call__(self, **kwargs):
        return await self.post(**kwargs)


internal = InternalRPC()
runtime = internal.runtime
rt = internal.rt
fn = internal.fn



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kess']

package_data = \
{'': ['*']}

install_requires = \
['dapr-ext-fastapi>=1.0.0,<2.0.0',
 'dapr>=1.0.0,<2.0.0',
 'fastapi>=0.63.0,<0.64.0']

setup_kwargs = {
    'name': 'kess',
    'version': '0.12.0',
    'description': '',
    'long_description': None,
    'author': 'majik',
    'author_email': 'me@yamajik.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os

FN_FOLDER = os.getenv("KESS_FN_FOLDER", "fn")
LIB_FOLDER = os.getenv("KESS_LIB_FOLDER", "lib")
RUNTIME_NAME  = os.getenv("KESS_RUNTIME_NAME", "")
RUNTIME_PREFIX = os.getenv("KESS_RUNTIME_PREFIX", "")
